"""
Base classes and framework for questionnaire objects

"""
import time, copy, re
from reportlab.lib import normalDate

_helpImgHTTPPath=None
_helpText='Help'
def setHelpImgHTTPPath(httpPath):
    global _helpImgHTTPPath
    _helpImgHTTPPath = httpPath

def setHelpText(text):
    global _helpText
    _helpText = text

class FieldValidationError(Exception):
    pass

def __dict_replace(s, d):
    """Replace substrings of a string using a dictionary."""
    for key, value in d.items():
        s = s.replace(key, value)
    return s

def unescape(data, entities={}):
    """Unescape &amp;, &lt;, and &gt; in a string of data.

    You can unescape other strings of data by passing a dictionary as
    the optional entities parameter.  The keys and values must all be
    strings; each key will be replaced with its corresponding value.
    """
    data = data.replace("&lt;", "<")
    data = data.replace("&gt;", ">")
    if entities:
        data = __dict_replace(data, entities)
    # must do ampersand last
    return data.replace("&amp;", "&")

def _deFixDict(dict):
    "Remove first prefix and return branched one"
    newDicts = {}
    newDicts[None] = {}
    for key, value in dict.items():
        chunks = key.split('.', 1)
        if len(chunks) == 1:
            newDicts[None][key] = value
        else:
            head, rest = chunks
            if not newDicts.has_key(head):
                newDicts[head] = {}
            newDicts[head][rest] = value
    return newDicts

def _getHeader(caption,tt=""):
    if tt:
        return """<tr><td colspan="2" height="19" class="TblRow1" width="70%%"><span class="TblHead7">%s</span></td><td width="30%%" align="right">%s</td></tr>\n""" % (caption,tt)
    else:
        return """<tr><td colspan="3" height="19" class="TblRow1"><span class="TblHead7">%s</span></td></tr>\n""" % caption

from reportlab.lib.utils import FmtSelfDict
class QBase(FmtSelfDict):
    def __init__(self):
        self.prefix = ''
        self.httpDir = None
        self.hidden = 0
        self.onLoad = ''
        self.onUnload = ''

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_parent']
        return state

    def dottedName(self):
        #return self.prefix + '.' + self.name
        return (self.prefix and self.prefix+'.' or '')+self.name

    def __getattr__(self, name):
        """Find an appropriate named child.

        This gets called if a lookup fails.  Thus, whatever is asked
        for is not an attribute.  See if it is the name of a child."""
        if self.__dict__.has_key('children'):
            for child in self.children:
                if child.name == name:
                    return child
        #still here?
        raise AttributeError("%s has no attribute or child named '%s'" % (self.__class__.__name__, name))

    def childNames(self):
        qBaseNames = []
        try:
            self.children
        except AttributeError:
            pass
        else:
            for child in self.children:
                try:
                    child.name
                except AttributeError:
                    pass
                else:
                    qBaseNames.append(child.name)
        return qBaseNames

    def deFixDict(self, params, qBaseNames=None):
        """As for _deFixDict(), but:

        Keys are returned for all qBases in the QDF tree, not only those in the
        HTTP dictionary.  For keys not in the HTTP dictionary, empty dictionary
        keys are returned.

        """
        if qBaseNames is None:
            qBaseNames = self.childNames()
        newParams = _deFixDict(params)
        for name in qBaseNames:
            newParams.setdefault(name, {})
        return newParams

    def getChildrenRecursively(self):
        """Return a list of all child QBase instances.

        Includes children, children of children, etc.

        """
        if not hasattr(self, 'children'):
            return []

        children = []
        for child in self.children:
            try:
                child.getChildrenRecursively
            except AttributeError:
                # some joker stuck a non-QBase in the tree
                continue
            children.append(child)
            children.extend([ch for ch in child.getChildrenRecursively()])
        return children

    def getParent(self):
        """Return immediate parent of this object in QDF tree."""
        return self._parent

    def getRoot(self):
        """Return root of the QDF tree (usually a Questionnaire instance)."""
        lastParent = self
        while 1:
            parent = lastParent._parent
            if parent is None:
                break
            lastParent = parent
        return lastParent

    def makeSample(self):
        "Initialize an object with sample data"
        if isinstance(self, QField) and not self.getTextValue():
            self.setTextValue(self.sampleData)
        if hasattr(self, 'children'):
            for child in self.children:
                if isinstance(child, QBase):
                    child.makeSample()

    def populateDict(self, dict, asText):
        "Do not modify dict in base class"
        pass

    def setTextDictionary(self, dict):
        """Unserialize QDF tree from HTTP parameters.

        This is called for every object in the QDF tree, regardless of whether
        HTTP parameters for this particular object were supplied.  XXX this may
        not always be true yet, due to use of self.deFixDict(D)[None].items():
        .deFixDict() only .setdefault()s the top-level keys in the dictionary
        it returns

        """
        pass

    def clear(self):
        "Reset all values to just-loaded state"
        pass
    def makeHtml(self):
        template = """
        <tr>
            <td colspan="3">%s</td>
        </tr>
        """
        return template % (self.prefix + '.' + self.name)

    def makeHiddenField(self):
        return '' #fields do stuff, base objects don't

    def getHtml(self):
        '''either makeHtml or makeHiddenField'''
        return (self.hidden and self.makeHiddenField or self.makeHtml)()

    def postInit(self):
        """Hack QDF tree.

        This hook is called on each QBase after entire QDF tree is loaded.

        """
        self.hidden = int(self.hidden)

    def fixup(self):
        """Hack QDF tree.

        This hook is called (by the root Questionnaire) on each QBase after
        entire QDF tree is loaded and populated.

        XXX currently only called for QSections

        """
        pass

    def visit(self,visitor):
        visitor(self)

    def findElementByName(self,name):
        return getattr(self,name,None)

    def getLText(self):
        self.lText = self.onLoad
        return self.lText

    def getUText(self):
        self.uText = self.onUnload
        return self.uText

class JsHideable:
    def _divMarker(self, start):
        return '<tr id="%s%s"%s />' % ((start and 'start' or 'end'), self.dottedName(), (self.jshidden and ' class="rlDivHidden"' or ''))
    def _divStartMarker(self): return self._divMarker(1)
    def _divEndMarker(self): return self._divMarker(0)

class QAddDivBase(QBase, JsHideable):
    def __init__(self):
        QBase.__init__(self)
        self.name = ''
        self.jshidden = 0

    def postInit(self):
        QBase.postInit(self)
        self.jshidden = int(self.jshidden)

class QAddStartDiv(QAddDivBase):
    def makeHtml(self):
        return self._divStartMarker()

class QAddEndDiv(QAddDivBase):
    def makeHtml(self):
        return self._divEndMarker()

class QHidden(QBase):
    def __init__(self):
        QBase.__init__(self)
        self.name = ''
        self.value = ''
        self.hidden = 1

    def setTextValue(self, newValue):
        "Accept string and do any needed conversions"
        if newValue in ('',None):
            self.value = None
        else:
            self.value = newValue

    def setTextDictionary(self, dict):
        if dict.has_key(self.name):
            self.setTextValue(dict[self.name])

    def getTextValue(self):
        if self.value is None:
            v = ''
        else:
            v = str(self.value)
        return v

    def makeHtml(self):
        html = []
        h = html.append
        h('<tr><td colspan="3">')
        h(self.makeHiddenField())
        h('</td></tr>')
        return '\n'.join(html)

    def makeHiddenField(self):
        template = """\t<input type="hidden" name="%s" value="%s">""" % (
            self.prefix+'.'+self.name,
            self.getTextValue()
            )
        return template

class QSeparator(QBase):
    def __init__(self):
        QBase.__init__(self)
        self.name = ''
        self.htmlClass = ''
        self.html = ''

    def makeHtml(self):
        html = self.html or "<hr>"
        htmlClass= self.htmlClass
        if htmlClass and 'class=' not in html: html.replace('>',' class="%s">'%htmlClass)
        name = self.name
        if name: html = '<span name="%s">%s</span>' % (name,html)
        return '<tr><td colspan="3">%s</td></tr>' % html

class QContainer(QBase):
    "has children - used for Questionnaire and Section"
    def __init__(self):
        QBase.__init__(self)
        self.label = None
        self.name = None
        self.children = []

    def populateDict(self, dict, asText):
        "Just ask child nodes...."
        for child in self.children:
            child.populateDict(dict, asText)

    def clear(self):
        for child in self.children:
            child.clear()

    def setTextDictionary(self, dict):
        """Set all element values from the dictionary.

        This is tricky for repeated groups as it may need
        to create on the fly.  Therefore instead of using
        Python's dynamic evaluation at the top we step down"""
        stuff = self.deFixDict(dict)
        # this is things without a prefix like 'age'
        myStuff = stuff[None]
        for key, value in myStuff.items():
            child = getattr(self, key, None)
            if child:
                child.setTextValue(value)

        # these go onto a child node like 'personalDetails.
        del stuff[None]
        for key, smallerDict in stuff.items():
            child = getattr(self, key, None)
            if child:
                child.setTextDictionary(smallerDict)

    def makeHtml(self):
        "Returns an HTML chunk as a form"
        if self.hidden: return ''
        prefix = self.prefix
        html = []
        if not int(getattr(self,'noCaption','0')):
            if self.label is not None:
                caption = self.label
            else:
                caption = prefix
            html.append(_getHeader(caption))
        for elem in self.children:
            html.append(elem.getHtml())
        return '\n'.join(html)

    def makeHiddenField(self):
        html = []
        for elem in self.children:
            html.append(elem.makeHiddenField())
        return '\n'.join(html)

    def visit(self,visitor):
        visitor(self)
        for c in self.children:
            if not getattr(c, 'hidden', 0):
                c.visit(visitor)

    def findElementByName(self, name):
        words = name.split('.', 1)
        if len(words) == 1:
            return getattr(self, name, None)
        else:
            child = getattr(self, words[0], None)
            if child: child = child.findElementByName(words[1])
            return child

class QSection(QContainer):
    "Represents a display section in a form"
    def __init__(self):
        QContainer.__init__(self)
        self.numbered = 1
        self.alias = None

    def postInit(self):
        QContainer.postInit(self)
        self.numbered = int(self.numbered)

    def getLabel(self):
        "Display to left of field"
        if self.label is None:
            return self.name
        else:
            return self.label

class QDivSection(QSection, JsHideable):
    "Section whose sole purpose in life is to group nodes so they can be hidden."
    def __init__(self):
        QSection.__init__(self)
        self.jshidden = 0
        self.numbered = 0

    def postInit(self):
        QBase.postInit(self)
        self.jshidden = int(self.jshidden)

    def getLabel(self):
        return ''

    def makeHtml(self):
        if self.hidden:
            return ''
        html = [self._divStartMarker()]
        for elem in self.children:
            html.append(elem.getHtml())
        html.append(self._divEndMarker())
        return '\n'.join(html)

class QField(QBase):
    "Base class for a field"
    def __init__(self):
        QBase.__init__(self)
        self.name = None
        self.editable = 1
        self.required = 0
        self.value = None
        self.label = None
        self.shortHelp = ''
        self.longHelp = ''
        self.maxlength = ''
        self.regex = None
        self.errorMessage = ''
        self.sampleData = ''
        self.onSubmit = ''
        self.onLoad = ''
        self.onUnload = ''
        self.jsEvents = ''
        self.jsHeader = ''
        self.rhsLabel = ''

    def __str__(self):
        "String representation should be field value"
        return self.getTextValue()

    def __repr__(self):
        if self.name is not None:
            namePart = ' (name %s)' % self.name
        else:
            namePart = ''
        return '<%s instance%s at %s>' % (self.__class__, namePart, id(self))

    def checkPattern(self, text):
        "Return true or raise error"
        if self.regex is None:
            return 1
        else:
            pat = re.compile(self.regex)
            match = pat.match(text)
            if match:
                # want it to match the whole string,
                # not just the front...
                if match.end() == len(text):
                    return 1
        #still here?  bomb out
        msg = self.getInputErrorMessage()
        raise FieldValidationError(msg)

    def getVText(self):
        self.vText = self.onSubmit
        return self.vText

    def getRText(self):
        fieldName = (self.prefix and self.prefix+'.' or '')+self.name
        regex = self.regex
        if regex or self.required=='1':
            regex = regex and ("/%s/" % regex) or 'null'
            msg = self.getInputErrorMessage(js=1)
            self.rText = "rl_checkRegex('%s', '%s', %s)" % (fieldName, msg, regex)
        else:
            self.rText = ''
        return self.rText

    def getLText(self):
        self.lText = self.onLoad
        return self.lText

    def getUText(self):
        self.uText = self.onUnload
        return self.uText

    def getIText(self):
        self.iText = self.jsHeader
        return self.iText

    def getInputErrorMessage(self,js=0):
        "Return the best error message possible"
        fieldName = self.label
        if self.errorMessage:
            msg = self.errorMessage
        elif self.regex:
            msg = "Please enter a valid value for '%s'" % fieldName
        elif self.required=='1':
            msg = 'Please enter a value for \'%s\'' % fieldName
        else:
            # XXX weird: is this req'd for backwards-compat?
            msg = ''
        if js: msg = msg.replace('\\','\\\\').replace("'", "\\'")
        return msg

    def setTextValue(self, newValue):
        "Accept string and do any needed conversions"
        if newValue in ('',None):
            self.value = None
        else:
            self.checkPattern(newValue)
            self.value = newValue

    def clear(self):
        self.value = None

    def setTextDictionary(self, dict):
        if dict.has_key(self.name):
            self.setTextValue(dict[self.name])

    def getTextValue(self):
        if self.value is None:
            v = ''
        else:
            v = str(self.value)
        return v

    def getLabel(self):
        "Display to left of field"
        if self.label is None:
            return self.name
        else:
            return self.label

    def getLongHelp(self):
        "Long Help as tooltip"
        if self.longHelp is not None:
            return self.longHelp
        else:
            return ''

    def getMaxLength(self):
        "use for max no. of chars"
        if self.maxlength is not None:
            return self.maxlength
        else:
            return ''

    def _getDotHelpEx(self,tt,sh):
        if tt:
            ttyoffset = getattr(self, 'ttyoffset', '')
            if ttyoffset:
                ttyoffs = ' ttyoffs="%s"' % ttyoffset
            else:
                ttyoffs = ''
            if _helpImgHTTPPath:
                tt = """<img align="top" src="%s" tooltip="%s" ttclass="tooltip"%s>""" % (_helpImgHTTPPath,tt,ttyoffs)
            else:
                tt = """<span tooltip="%s" ttclass="tooltip" class="tttext"%s>%s</span>""" % (tt,ttyoffs,_helpText)
            sh = sh and ('%s&nbsp;&nbsp;%s' % (sh,tt)) or tt
        return sh

    def _getDotHelp(self):
        return self._getDotHelpEx(self.getLongHelp(), self.shortHelp)

    def makeHtml(self, _jsEvents=''):
        html = []
        h = html.append
        h('<tr><td width="30%%">%s</td>' % self.getLabel())
        h('<td width="40%%">')
        if self.editable:
            h('<input type="text" name="%s.%s" maxlength="%s" value="%s" %s/>%s' %
              (self.prefix,
               self.name,
               self.getMaxLength(),
               self.getTextValue(),
               ' '.join([self.jsEvents, _jsEvents]),
               self.rhsLabel,
               )
              )
            h('</td>')
            h('<td width="30%%" align="right">%s</td>' % self._getDotHelp())
        else:
            h('%s</td>' % self.getTextValue())
            h(self.makeHiddenField())
            h('<td width="40%%"></td>')
        h('</tr>')
        return '\n'.join(html)

    def makeHiddenField(self):
        from rlextra.utils.cgisupport import quoteValue
        V = self.getTextValue()
        if not V:
            result = ''
        else:
            fmt = """\t<input type="hidden" name="%s" value="%s">"""
            name = self.prefix+'.'+self.name
            # hack: must quoteValue() to HTML-quote strings originating internally to webeforms
            # (even though all CGI params are quoteValue()d on input)
            if type(V) in (type(()),type([])):
                name += ':list'
                result = ''.join([fmt % (name,quoteValue(v)) for v in V])
            else:
                result = fmt % (name,quoteValue(V))
        return result

    def populateDict(self, dict, asText):
        "Just ask child nodes...."
        if self.value in [None, '']:
            pass
        key = self.prefix+'.'+self.name
        if asText:
            dict[key] = str(self.value)
        else:
            dict[key] = self.value


class QTextField(QField):
    pass


class QFloatField(QField):
    _typeName = 'FLOAT'
    def __init__(self):
        QField.__init__(self)
        self.min = None
        self.max = None
        self.required = 0

    def postInit(self):
        QField.postInit(self)
        self.widname = self.name.replace('.','_')
        self.id = '%s.%s' % (self.prefix, self.name)

    def makeHtml(self):
        return QField.makeHtml(self, _jsEvents=self._fmt('onChange="%(widname)s.changed()"'))

    def getIText(self):
        iText = []
        min = self.min is None and 'null' or self.min
        max = self.max is None and 'null' or self.max
        entitydefs = {'&pound;': '\xa3'}
        label = unescape(self.label, entitydefs).replace('\\','\\\\').replace("'","\\'")
        iText.append("var %s = new RL_%s('%s',%s,%s,'%s',%d);" % (self.widname, self._typeName, self.id, min, max, label, int(self.required)))
        self.iText = '\n'.join(iText)
        if self.jsHeader: self.iText += '\n'+self.jsHeader+';'
        return self.iText

    def getVText(self):
        self.vText = self.onSubmit or self._fmt('%(widname)s.validate()')
        return self.vText

    def setTextValue(self, newValue):
        "Accept string and do any needed conversions"
        if newValue:
            self.value = float(newValue)

    def getRText(self):
        self.rText = ''
        return self.rText

class QIntegerField(QField):
    _typeName = 'INT'
    errorMessage = 'Please supply a whole number (no decimal point)'

    def setTextValue(self, newValue):
        "Accept string and do any needed conversions"
        if newValue:
            self.value = int(newValue)

class QDateField(QField):
    """For now accept UK dates only - any of 31/12/2003,
    31.12.2003, 31-12-2003, 31-12-03"""
    def __init__(self):
        QField.__init__(self)
        #dd/mm/yyyy or dd-mm-yyyy
        self.regex = r"\d\d?(-|\/)\d\d?(-|\/)\d\d\d\d"
        self.errorMessage = 'Please use format dd/mm/yyyy e.g. 31/12/2002'

    def parseDate(self, dateStr):
        dd = mmm = yyyy = None
        for delim in '-/.':
            if len(dateStr.split(delim))==3:
                dd,mmm,yyyy = dateStr.split(delim)
        if dd == None:
            raise ValueError('bad date value "%s", use e.g. "23/01/2003"' % dateStr)
        yyyy = int(yyyy)
        if yyyy < 1000:
            if yyyy < 25:  # pivot date
                yyyy = yyyy + 2000
            else:
                yyyy = yyyy + 1900

        nd = normalDate.NormalDate((yyyy,int(mmm),int(dd)))
        return nd

    def setTextValue(self, newValue):
        if newValue == '':
            self.value = None
        else:
            self.value = self.parseDate(newValue)

    def getTextValue(self):
        if self.value is None:
            return ''
        else:
            return self.value.formatMS('{dd}-{mm}-{yyyy}')

class QBooleanField(QField):
    def __init__(self):
        QField.__init__(self)
        self.spanAllCols = 'no'

    def setTextValue(self, newValue):
        "Accept string and do any needed conversions"
        if newValue in ['on', #web forms do this,
                        '1',
                        'yes']:
            self.value = 1
        else:
            self.value = 0

    def makeHtml(self, _attrStrings=[]):
        if self.spanAllCols == 'yes':
            colspan = ' colspan="3"'
            labelTd = ''
            inputLabel = self.getLabel()+'&nbsp;&nbsp;&nbsp;'
        else:
            colspan = ''
            labelTd = '<td>%s</td>' % self.getLabel()
            inputLabel = ''
        xtras = _attrStrings+(self.value and ['checked'] or [])+(self.jsEvents and [self.jsEvents] or [])
        return '<tr>%s<td%s>%s<input type="checkbox" name="%s.%s"%s/></td><td>%s</td></tr>' % (
                            labelTd,
                            colspan,
                            inputLabel,
                            self.prefix,
                            self.name,
                            xtras and ' '.join(['']+xtras) or '',
                            self._getDotHelp()
                            )

class QConfirmField(QBooleanField):

    def __init__(self):
        QField.__init__(self)
        self.spanAllCols = 'no'
        self.errorMessage = 'Please check the confirmation checkbox'
        self.target = None

    def makeHtml(self):
        html = []
        a = html.append
        if self.target is not None:
            onClick = r"getElementWithId('%s').style.display = this.checked ? 'block' : 'none';" % (
                self.target,)
            attrStrings = ['onclick=%s' % onClick]
        else:
            attrStrings = []
        a(QBooleanField.makeHtml(self, attrStrings))
        return '\n'.join(html)

    def getVText(self):
        value = "document.forms[0].elements['%s'].checked" % self.dottedName()
        msg = self.getInputErrorMessage(js=1)
        confirm = "(%s?'':'%s')" % (value, msg)
        vText = [confirm]
        if self.onSubmit:
            vText.append(self.onSubmit)
        self.vText = '+"\\n"+'.join(vText)
        return self.vText

class QText(QField):
    def __init__(self):
        QField.__init__(self)
        self.name = None
        self.label = None
        self.value = None

    def makeHtml(self):
        label = self.getLabel()
        tt = self._getDotHelp()
        if tt:
            return '<tr><td colspan=2>' + (label or '&nbsp;') + '</td><td width="30%%" align="right">' + tt + '</td></tr>'
        else:
            return '<tr><td colspan=3>' + (label or '&nbsp;') + '</td></tr>'

class QHLink(QField):
    def __init__(self):
        QField.__init__(self)
        self.name = None
        self.label = None
        self.hlink = None
        self.displayed = 1
        self.confirm = None

    def getHlink(self):
        if self.hlink is not None:
            return self.hlink
        else:
            return ''

    def makeHtml(self):
        label = self.getLabel()
        html = []
        a = html.append
        if not self.displayed:
            trStyle =' style="display:none"'
            trId = ''
        else:
            if self.confirm is not None:
                onSubmit = r"getElementWithId('%s').style.display = this.checked ? 'block' : 'none';" % (
                    self.dottedName(),)
                a('<tr><td colspan="3"><input type="checkbox" onclick="%s">%s</td></tr>' % (onSubmit, self.confirm))
                trId = ' id="%s"' % self.dottedName()
                trStyle =' style="display:none"'
            else:
                trStyle = trId = ''
        a('<tr><td colspan=3><span%s%s><a href="%s">%s</a></span></td></tr>' % (trId, trStyle, self.getHlink(), (label or '&nbsp;')))
        return '\n'.join(html)

class QHeader(QText):
    def makeHtml(self):
        return _getHeader(self.getLabel() or '&nbsp;',tt=self._getDotHelp())

_oc_re = re.compile(r'''\s+onchange\s*=\s*(?:['"]|)''',re.I)  #"
class QChoice(QField):
    def __init__(self):
        QField.__init__(self)
        self._initialized = 0
        self.name = None
        self.value = None
        self.label = None
        self.spanAllCols = 'no'
        self.shortHelp = ''
        self.options = []
        self.labels = []
        self.style = 'dropdown' # also have 'radio', 'multidropdown' & 'multibox'
        self.ttyoffset = ""
        self.required = 0
        self.errorMessage = ''
        self.forceFirstOn = '1'
        self.other = ''
        self.otherSize = ''
        self.otherMax = ''

        self.makeDiv = ''
        self.divName = ''

    def postInit(self):
        QField.postInit(self)
        #called AFTER options have been set
        self._initialized = 1

    def removeOption(self, name):
        """To be called at fixup() time."""
        ii = self.options.index(name)
        self.options.remove(name)
        del self.labels[ii]

    def removeAllOptions(self):
        """To be called at fixup() time."""
        for name in list(self.options):
            self.removeOption(name)

    def addOptions(self, options):
        """To be called at fixup() time."""
        for optionCode, label in options:
            self.options.append(optionCode)
            self.labels.append(label)

    def addChild(self, child):
        label = child.label
        value = child.value
        self.options.append(value)
        if label is None:
            self.labels.append(value)
        else:
            self.labels.append(label)

    def getTextValue(self):
        if self.style.startswith('multi'):
            V = self.value
            if not V: V = []
            else:
                if type(V) not in (tuple,list): V = [V]
                return [str(v) for v in V]
        else:
            return QField.getTextValue(self)

    def setTextValue(self, newValue):
        if self._initialized: #validate if options set
            options = self.options
            if self.style.startswith('multi'):
                if not newValue: newValue = []
                elif not self.other:
                    if type(newValue) not in (tuple,list):
                        newValue = [newValue]
                        invalid = [v for v in newValue if v not in options]
                        if invalid:
                            raise KeyError('%s not in allowed values %s' % (invalid, options))

            else:
                if not newValue:  # must allow empty string initializing
                    newValue = None
                elif not self.other and newValue not in options:
                    raise KeyError('%s not in allowed values %s' % (newValue, options))
        self.value = newValue

    def makeHtml(self):
        if not self.editable:
            return self.makeStaticText()
        elif self.style in ('dropdown','multidropdown'):
            if self.makeDiv:
                html = []
                h = html.append
                h("</table><div id='%s'><table>" % self.divName)
                h(self.makeHtmlCombo())
                h("</table></div><table>")
                return '\n'.join(html)
            else:
                return self.makeHtmlCombo()
        elif self.style in ('radiobutton','multibox'):
            return self.makeHtmlRadios()
        else:
            return ''

    def makeStaticText(self):
        html = []
        h = html.append
        h('<tr><td>%s</td>' % self.getLabel())
        h('<td>%s</td>' % self.getTextValue())
        h(self.makeHiddenField())
        h('<td></td></tr>')
        return '\n'.join(html)

    def makeHtmlCombo(self):
        label = self.getLabel()
        labelTd = '<td>%s</td>' % label
        if self.spanAllCols == 'yes':
            dotHelp = ''
            if not label:
                labelTd = ''
                colspan = ' colspan="3"'
            else:
                colspan = ' colspan="2"'
        else:
            colspan = ''
            dotHelp = '<td align="right">%s</td>' % self._getDotHelp()
        value = self.value
        options = self.options
        labels = self.labels
        pfxname = self.prefix+'.'+self.name
        jsEvents = self.jsEvents and (' '+self.jsEvents) or ''
        style = self.style
        other = self.other
        if other:
            assert style!='multidropdown','other not allowed with multidropdown'
            m = _oc_re.search(jsEvents)
            if m:
                j = m.end()
                jsEvents = jsEvents[:j]+'RL_OTHERDD_ONCHANGE(this);'+jsEvents[j:]
            else:
                jsEvents = ' onchange="RL_OTHERDD_ONCHANGE(this)"'+jsEvents
            ddsfx = ':dd'
            hidVal = value
            if value and value not in options:
                othVal = value
                value = '*other*'
            else:
                othVal = ''
            options = options+['*other*']
            labels = labels+[other]
        else:
            ddsfx = ''
        if style=='multidropdown':
            pfxname += ':list'
            J = [j for j,v in enumerate(options) if v in (value or [])]
            jsEvents += 'multiple '
        else:
            J = [value not in options and -1 or options.index(value)]
        html = ['''<tr>%(labelTd)s<td%(colspan)s><select name="%(pfxname)s%(ddsfx)s"%(jsEvents)s>''' \
                    % locals()]
        for i in xrange(len(options)):
            html.append('    <option value="%s"%s>%s</option>' %
                        (options[i], i in J and " selected" or "", labels[i])
                        )
        html.append("""</select></td>%s</tr>\n"""% dotHelp)
        if other:
            otherSize = self.otherSize or "5"
            otherMax = self.otherMax or "30"
            labelTd = (not colspan and '<td></td>' or '')
            html.append('''<tr>%(labelTd)s<td%(colspan)s>
                                <input type="text" name="%(pfxname)s:eb" size="%(otherSize)s" maxlength="%(otherMax)s" value="%(othVal)s"/>
                                <input type=hidden name=%(pfxname)s value="%(hidVal)s"/><sup>(%(other)s)</sup></td>%(labelTd)s</tr>''' % locals())
        return '\n'.join(html)

    def makeHtmlRadios(self):
        helpText = self._getDotHelp()
        horizMax = 2
        if not helpText:
            colspan = ' colspan="2"'
        else:
            colspan = ''
        if self.spanAllCols == 'yes':
            colspan = ' colspan="3"'
            horizMax = 3
            labelTd = ''
        else:
            labelTd = '<td>%s</td>' % self.getLabel()
        html = ['<tr>%s<td%s><p>' % (labelTd, colspan)]
        value = self.value
        options = self.options
        n = len(options)
        name = self.name

        layout = getattr(self,'layout','countswitch')
        if layout=='countswitch':
            layout = n<=horizMax and 'horizontal' or 'vertical'

        assert not self.other,'other not allowed with choice boxes'
        if self.style=='multibox':
            name += ':list'
            J = [j for j,v in enumerate(options) if v in (value or [])]
            inputType = 'checkbox'
        else:
            inputType = 'radio'
        if value not in options:
            J = self.forceFirstOn=='1' and [0] or []
        else:
            j = [options.index(value)]
        jsEvents = self.jsEvents and (' '+self.jsEvents) or ''
        for i in xrange(n):
            checked = i in J and " checked" or ""
            v = '<input type="%s" name="%s.%s" value="%s"%s%s>' \
                    % (inputType,self.prefix, name, options[i], checked, jsEvents)
            if layout=='vertical':
                v += self.labels[i]+'<br>'  #ie box label<br>box label<br> ...
            else:
                v = self.labels[i]+v        #ie label box label box ...
                if layout=='matrix': v += not((i+1)%horizMax) and '<br>' or ''
            html.append(v)
        html.append("""</p></td><td align="right">%s</td></tr>\n""" % helpText)
        return '\n'.join(html)

    def getRText(self):
        if not self.other and self.required=='1':
            name = (self.prefix and self.prefix+'.' or '')+self.name
            if self.style.startswith('multi'): name += ':list'
            msg = self.getInputErrorMessage(js=1)
            rstyle = self.style.endswith('dropdown') and 'single' or 'multi'
            self.rText = '''RL_%s_choice_Required('%s','%s')''' % (rstyle,name,msg)
        else:
            self.rText = ''
        return self.rText

    def getVText(self):
        other = self.other
        if other:
            self.vText = "RL_OTHERDD_ONSUBMIT('%s.%s','%s')" % (self.prefix,self.name,(self.required=='1' and self.errorMessage or ''))
        else:
            self.vText = ''
        return self.vText

    def getLText(self):
        if self.other:
            self.lText = 'RL_OTHERDD_ONLOAD("%s.%s")' % (self.prefix,self.name)
        else:
            self.lText = ''
        statements = []
        if self.onLoad:
            statements.append(self.onLoad) 
        if self.lText:
            statements.append(self.lText) 
        return '; '.join(statements) 

    def getHText(self): 
        from qwidget import getJSFiles
        L = []
        if self.other: L += ['<script language="Javascript" type="text/javascript" src="%(httpDir)s/rl_qbase.js"></script>']
        self.hText = getJSFiles(L,fmt=self._fmt)
        return self.hText

class QRepeatedGroup(QContainer):
    """This can create multiple rows based on a 'pattern'.

    example:  holds between 1 and 7 groupings of
      (fundCode, fundName, amount)
    i.e. 2 text fields and a floatField.

    The self.substructure field is populated at parse time
    with one example each of the type of fields allowed.
    The things in the group MUST exhibit the same structure
    in each row.  At runtime it generates a list of children,
    each of which is a QSection with named children, such
    as the 3 above.  To append a row to the children list
    (i.e. create a new fundCode/name.amount triple), call
    addRow().
    """
    def __init__(self):
        QContainer.__init__(self)
        self.minCount = 0 # zero allowed
        self.maxCount = None  # i.e. no maximum
        self.sampleData = 0
        self.substructure = []
        self._autoCreateRows = 0

    def postInit(self):
        "called after last child added"
        QContainer.postInit(self)
        if self.maxCount:
            self.maxCount = int(self.maxCount)
        self.sampleData = int(self.sampleData)

    def makeSample(self):
        "Initialize an object with sample data"
        # XXX I'm not sure if this is still correct: both this method and
        # setTextDictionary are now called when using sample data, so perhaps
        # addRow is not called correct nr of times
        for i in range(self.sampleData):
            self.addRow()
        for child in self.children:
            child.makeSample()

    def clear(self):
        self.children = []

    def addChild(self, child):
        self.substructure.append(child)

    def addRow(self):
        sec = QSection()
        sec.noCaption = 1
        sec.name = '%s.row%d' % (self.name,len(self.children))
        for thing in self.substructure:
            # Andy's algorithm is horribly broken for the case
            # when the substructure items are complex and have
            # dictionaries/lists as attributes. A step too far
            # in greedy optimization
            # this copy algorithm makes one from a prototype
            # and is 20x faster than using copy.copy or
            # copy.deepcopy  !?!??
            #klass = thing.__class__
            #thang = klass()
            #thang.__dict__.update(thing.__dict__)
            #thang.prefix = self.prefix + '.' + sec.name

            # use deepcopy, it ensures that we get guaranteed different
            # objects
            # we should really do the thing in three stages.
            # 1) create 2) apply attributes 3) do the postInit
            # but we have no proper way to get the initial attributes
            # from the template objects
            thang = copy.deepcopy(thing)
            thang._parent = sec
            thang.prefix = self.prefix + '.' + sec.name
            thang.postInit()
            sec.children.append(thang)
        self.children.append(sec)
        return sec

    def deleteRow(self):
        "Deletes last row"
        self.children = self.children[:-1]

    def setTextDictionary(self, dict):
        self._autoCreateRows = 1
        QContainer.setTextDictionary(self, dict)
        self._autoCreateRows = 0

    def __getattr__(self, name):
        "If they ask for row4, make sure 4 rows exist..."

        if name[0:3] == 'row':
            rowNo = int(name[3:])
            if rowNo < len(self.children):
                return self.children[rowNo]
            elif self._autoCreateRows:
                while len(self.children) < rowNo+1:
                    self.addRow()
                return self.children[rowNo]
            else:
                raise IndexError('QRepeatedGroup does not have enough rows')
        else:
            raise AttributeError('Attribute %s not found in QRepeatedGroup' % name)

    def __getitem__(self, idx):
        return self.children[idx]

    def getLabel(self):
        "Display to left of field"
        if self.label is None:
            return self.name
        else:
            return self.label

    def makeHtml(self):
        """Must return as many rows as asked for..."""
        html = []
        rowLabel = self.getLabel()
        if rowLabel:
            rowLabelFmt = """%s %%d<!--%%s-->\n""" % rowLabel
        # we might have 3 rows but allow up to 7.  Generate more
        # rows if needed so stuff can be input.
        childCount = len(self.children)
        if self.maxCount:
            extraNeeded = self.maxCount - childCount
            for i in range(extraNeeded):
                self.addRow()

        html.append('<!-- QRepeatedGroup name="%s" prefix="%s" -->'%(self.name, self.prefix))
        for elem in self.children:
            html.append(elem.getHtml())

        if self.maxCount:
            for i in range(extraNeeded):
                self.deleteRow()

        return '\n'.join(html)
