#Copyright ReportLab Europe Ltd. 2000-2005
#see license.txt for license details
__version__=''' $Id: fieldreader.py 20653 2006-03-24 15:03:33Z jjlee $ '''
__all__=['FieldReader','XMLTreeGetter']
from rml2pdf import readLength
def _defaultDataGetter(name, intent=None):
    return '''<!-- unknown data for field %s -->''' % name

def strStrip(s):
    return s.strip()


class XMLTreeGetter:
    '''
    >>> from rlextra.radxml import xmlutils
    >>> T = xmlutils.xml2doctree('<a><b>bval</b><c ca="cattr"><d>dval</d></c></a>',0)
    >>> x = XMLTreeGetter(T,stripDots=1)
    >>> print x('a.b')
    bval
    >>> print x('a.c.d')
    dval
    >>> print x('a.c.ca')
    cattr
    '''
    def __init__(self,tree,stripDots=0,default='',prefix='',strip=lambda x: x):
        self.tree = tree
        self.stripDots = stripDots
        self.default = ''
        self.prefix = prefix
        self.strip = strip

    def getValue(self,name,intent=None):
        name = self.prefix+name
        A = name.split('.')
        if self.stripDots: A = A[self.stripDots:]
        obj = self.tree
        for a in A:
            obj = getattr(obj,a)
        return self.strip(str(obj))

    def __call__(self,name,intent=None):
        try:
            return self.getValue(name,intent)
        except:
            if self.default is not None: return self.default
            raise

class FieldsGetter:
    """Get a sequence of positioned Fields whose text comes from a single data field.

    You pass instances satisfying this interface (e.g. this class, or more
    usually a subclass, for convenience) to FieldReader.multiTextRml() -- see
    that docstring for what this stuff is for.

    Interface for FieldGetter instances:

    .setFieldMap() will be called with a mapping of field name to field
    objects.  The field objects have attributes as returned by dictFromRow().

    The instance will be called as a callable with field name and content
    and should return a list of Fields.

    The base class also provides an implementation.  The implementation assumes
    the (fixed number, nrParts, of) positioned field names to be used to are
    generated by appending integer indices to the name passed in when calling
    the instance, e.g. 'postcode' --> 'postcode1', 'postcode2'.  To split the
    data into chunks, it uses a normalization function and a separator
    character.

    """
    def __init__(self, normalize, nrParts, separator):
        self.nrParts = nrParts
        self.separator = separator
        self.normalize = normalize

    def setFieldMap(self, fieldFromName):
        self.fieldFromName = fieldFromName

    def __call__(self, name, content):
        content = self.normalize(content)
        parts = content.split(' ')
        assert len(parts) == self.nrParts
        fields = []
        for ii, part in enumerate(parts):
            partName = '%s%d' % (name, ii+1)
            field = self.fieldFromName[partName]
            field.content = part
            fields.append(field)
        return fields


class PostcodeFieldsGetter(FieldsGetter):
    def __init__(self):
        from rlextra.utils.layoututils import postcodeFormatter
        pcf = lambda t: postcodeFormatter(t, singleSpace=True)
        FieldsGetter.__init__(self, pcf, nrParts=2, separator=' ')

class IntentSpec:
    def __init__(self,tagName,valueName,attributeNames,isEmpty=False,preRenderFunc=None):
        self.tagName = tagName
        self.valueName = valueName
        self.attributeNames = attributeNames
        self.isEmpty = isEmpty
        self.preRenderFunc = preRenderFunc

class FixBoolean:
    def __init__(self,name):
        self.name = name
    def __call__(self,D):
        value = D[self.name]
        if not value:
            D[self.name] = '0'

class Field:
    def __init__(self, pairs):
        self.__dict__.update(dict(pairs))
##     def __str__(self):
##         r = []
##         for kv in self.__dict__.items():
##             r.append('%s=%r' % kv)
##         return '<%s %s>' % (self.__class__, ', '.join(r))

FIELD_INFO_NAMES = ['name', 'pageNum', 'x', 'y', 'width', 'height', 'intent', 'count']
       
def dictFromRow(row):
    nkwds = {}
    for ii, name in enumerate(FIELD_INFO_NAMES):
        nkwds[name] = row[ii]
    nkwds['name'] = nkwds['name'].strip()
    return nkwds


class FieldReader:
    '''
    >>> from StringIO import StringIO
    >>> csv=StringIO("""name,pageNum,x,y,width,height,intent\\nsippDetails.clientDetails.houseName,3,21.0,604.0,352.0,16.0,letterBoxes,10\\nsippDetails.clientDetails.occupation,3,21.0,410.0,552.0,16.0,textBox,0\\nsippDetails.clientDetails.isExistingClient,3,23.0,499.0,10.5,9.5,checkBox,0\\nsippDetails.clientDetails.amount,3,20,100,30,10,amount,3""")
    >>> F = FieldReader(csv)
    >>> print F.fieldRml('sippDetails.clientDetails.houseName',content='Hello World')
    <letterBoxes boxHeight='16.0' boxWidth='35.2' count='10' x='21.0' y='604.0'>Hello World</letterBoxes>
    >>> print F.fieldRml('sippDetails.clientDetails.occupation',content='Widget Grokker')
    <textBox boxHeight='16.0' boxWidth='552.0' x='21.0' y='410.0'>Widget Grokker</textBox>
    >>> print F.fieldRml('sippDetails.clientDetails.isExistingClient',checked="1")
    <checkBox boxHeight='9.5' boxWidth='10.5' checked='1' x='23.0' y='499.0'/>
    >>> print F.fieldRml('sippDetails.clientDetails.amount',content="123.454")
    <letterBoxes boxHeight='10' boxWidth='10.0' count='5' x='0.0' y='100'>12345</letterBoxes>
    >>> print F.fieldRml('sippDetails.clientDetails.amount',content="1234.56")
    <letterBoxes boxHeight='10' boxWidth='10.0' count='6' x='-10.0' y='100'>123456</letterBoxes>

    We can also use the xmltree as a value getter
    >>> csv.seek(0)
    >>> from rlextra.radxml import xmlutils
    >>> T=xmlutils.xml2doctree('<a><clientDetails houseName="The Willows" occupation="Builder"/></a>',0)
    >>> F=FieldReader(csv,XMLTreeGetter(T,stripDots=1))
    >>> print F.fieldRml('sippDetails.clientDetails.houseName')
    <letterBoxes boxHeight='16.0' boxWidth='35.2' count='10' x='21.0' y='604.0'>The Willows</letterBoxes>
    >>> print F.fieldRml('sippDetails.clientDetails.occupation')
    <textBox boxHeight='16.0' boxWidth='552.0' x='21.0' y='410.0'>Builder</textBox>

    playing with defaults
    >>> F.setDefaults(textBox=dict(textColor='red'))
    >>> print F.fieldRml('sippDetails.clientDetails.occupation')
    <textBox boxHeight='16.0' boxWidth='552.0' textColor='red' x='21.0' y='410.0'>Builder</textBox>
    >>> F.setDefaults(textBox=dict(textColor=None,shrinkToFit='1'))
    >>> print F.fieldRml('sippDetails.clientDetails.occupation')
    <textBox boxHeight='16.0' boxWidth='552.0' shrinkToFit='1' x='21.0' y='410.0'>Builder</textBox>
    '''
    def __init__(self,fileName,dataGetter=_defaultDataGetter,defaults={}):
        '''
        fileName        the .fld file
        dataGetter      a callable to get default content using the field name
        defaults        dict of dict of default values ie defaults[intent][attributeName]
                        will be used if set and not overriden
        '''
        self.fileName = fileName
        self.dataGetter = dataGetter
        self.FIELD_ROWS = self.readFields(fileName)
        self._defaults = {}
        self._defaults['letterBoxes'] = self._defaults['letterBox'] = {}    #hack aliasing
        self.setDefaults(**defaults)

    def setDefaults(self,**defaults):
        for intent,intentDefaults in defaults.items():
            if not self.intentSpecs.has_key(intent):
                raise KeyError("Don't know about intent \"%s\"" % intent)
            spec = self.intentSpecs[intent]
            D = self._defaults.setdefault(intent,{})
            for kk,vv in intentDefaults.items():
                if kk not in spec.attributeNames:
                    raise ValueError('Intent "%s" has no attribute named "%s"' % (intent,kk))
                if vv is None:
                    if D.has_key(kk): del D[kk]
                else:
                    D[kk] = vv

    def readFields(self,fileName):
        import csv
        from reportlab.lib.utils import open_for_read
        if hasattr(fileName,'read'):
            f=fileName
        else:
            f=open_for_read(fileName, "rb")
        ROWS=[map(str.strip,row) for row in csv.reader(f)][1:]
        if f is not fileName: f.close()
        return ROWS

    def _render(self,name,intent,dataGetter,**kwds):
        from xml.sax.saxutils import escape
        spec = self.intentSpecs[intent]
        valueName = spec.valueName
        D=self._defaults.get(intent,{}).copy()
        escapeAction=D.get('escape')
        D.update(kwds)
        if not kwds.has_key(valueName):
            data = dataGetter(name, intent)
            if escapeAction == 'escape':
                data = escape(data)
            D[valueName] = data
        if spec.preRenderFunc: spec.preRenderFunc(D)
        rml = []
        add = rml.append
        if D.has_key('rotation'):
            angle = D['rotation']
            add('<saveState/><rotate degrees="%s"/>' % angle)
            angle = float(angle)
            from reportlab.graphics.shapes import rotate, inverse, transformPoint
            D['x'],D['y'] = transformPoint(inverse(rotate(float(angle))),(float(D['x']),float(D['y'])))
        attrs = []
        aa = attrs.append
        aNames = spec.attributeNames
        for k,v in D.items():
            if k in aNames:
                aa(' %s=%s' % (k,repr(str(v))))
        attrs.sort()
        d = dict(tagName=spec.tagName,attrs=''.join(attrs))
        if spec.isEmpty:
            add('<%(tagName)s%(attrs)s/>' % d)
        else:
            d['content'] = D[valueName]
            add('<%(tagName)s%(attrs)s>%(content)s</%(tagName)s>' % d)
        if D.has_key('rotation'): add('<restoreState/>')
        return ''.join(rml)#+('<!--%s-->' % name)

    def _fieldRml(self,dataGetter,name,pageNum,width,height,intent,count,**kwds):
        if intent in ('letterBoxes','letterBox'):
            kwds['count']=count
            kwds.setdefault('boxWidth',str(readLength(width)/int(count)))
            kwds.setdefault('boxHeight',height)
        elif intent=='checkBox':
            kwds.setdefault('boxWidth',width)
            kwds.setdefault('boxHeight',height)
        elif intent=='textBox':
            kwds.setdefault('boxWidth',width)
            kwds.setdefault('boxHeight',height)
        elif intent=='amount':
            count = int(count)
            ndp = count - 1
            kwds.setdefault('boxHeight',height)
            boxWidth = float(kwds.setdefault('boxWidth',str(readLength(width)/count)))
            content = float(kwds['content'])
            kwds['content'] = text = ('%%.%dd' % count) % int(content*10**ndp)
            x = readLength(kwds['x'])
            x -= (len(text)-count)*boxWidth
            kwds['x'] = str(x)
            kwds['count'] = len(text)
        else:
            return '''<!-- unknown intent "%(name)s" "%(pageNum)s" "%(width)s" "%(height)s" "%(intent)s" "%(count)s" "%(kwds)s" -->''' % locals()
        return self._render(name,intent,dataGetter,**kwds)

    def multiTextRml(self,name,fieldsGetter,dataGetter=None,debugLabelDataGetter=None,**kwds):
        """Return RML for things like postcodes, National Insurance numbers.

        This is for formatting single-data-field text that is displayed using
        more than one positioned field (usually using letterBox).  For example:
        postcodes, NI numbers sort codes, etc.

        We want to give one name in a prep file and have the code fetch all the
        fields we need, get the content for each, and format the content in
        blocks.

        fieldsGetter is an object satisfying FieldsGetter interface.

        """
        if dataGetter is None:
            dataGetter = self.dataGetter
        rowDict = dict([(row[0], Field(dictFromRow(row))) for row in self.FIELD_ROWS])
        fieldsGetter.setFieldMap(rowDict)
        content = dataGetter(name)
        fields = fieldsGetter(name, content)
        rml = []
        for field in fields:
            kwds['content'] = field.content
            rml.append(self.fieldRml(field.name,dataGetter,**kwds))
        return '\n'.join(rml)

    def fieldSetRml(self,pattern,dataGetter,debugLabelDataGetter=None,**kwds):
        import re
        import pprint
        fields = [row for row in self.FIELD_ROWS if re.search(pattern, row[0])]
        return self._fieldSetRml(fields,kwds,dataGetter,debugLabelDataGetter)

    def fieldRml(self,name,dataGetter=None,**kwds):
        from xml.sax.saxutils import escape
        rows=[row for row in self.FIELD_ROWS if row[0].strip()==name]
        if not rows:
            def htmlCommentEscape(text):
                return escape(text.replace('--', '__'))
            return ('''<!--RML for %s kwds=%s should have been here-->''' %
                    (htmlCommentEscape(name), htmlCommentEscape(str(kwds))))
        return self._fieldSetRml(rows,kwds,dataGetter)

    def coords(self,name,dataGetter=None):
        rows=[row for row in self.FIELD_ROWS if row[0].strip()==name]
        if len(rows) != 1:
            raise ValueError()
        row = rows[0]
        rowDict = dictFromRow(row)
        return tuple([float(c) for c in rowDict['x'], rowDict['y']])

    def _fieldSetRml(self,rows,kwds,dataGetter=None,debugLabelDataGetter=None):
        if dataGetter is None:
            dataGetter = self.dataGetter
        rml = []
        for ROW in rows:
            nkwds = dictFromRow(ROW)
            nkwds.update(kwds)
            name = nkwds['name']
            pageNum = nkwds['pageNum']
            width=nkwds['width']
            height=nkwds['height']
            intent=nkwds['intent']
            count=nkwds['count']
            for unwanted in ('name','pageNum',
                'width','height','intent','count'):
                del nkwds[unwanted]
            rml.append(self._fieldRml(dataGetter,name,pageNum,width,height,intent,count,**nkwds))
            if debugLabelDataGetter is not None:
                x = float(nkwds['x'])
                y = float(nkwds['y'])
                boxWidth = 100
                boxHeight = 7
                rml.append(self._render(name,'textBox',debugLabelDataGetter,
                                        x=str(x),y=str(y),
                                        boxWidth=str(boxWidth),
                                        boxHeight=str(boxHeight),
                                        #boxFillColor='white',
                                        )
                           )
        return '\n'.join(rml)

    intentSpecs = dict(
            checkBox=IntentSpec('checkBox','checked','''style x y labelFontName labelFontSize labelTextColor boxWidth boxHeight checkStrokeColor boxStrokeColor boxFillColor lineWidth line1 line2 line3 checked bold graphicOn graphicOff'''.split(),True,preRenderFunc=FixBoolean('checked')),
            textBox=IntentSpec('textBox','content','''style x y boxWidth boxHeight labelFontName labelFontSize labelTextColor labelOffsetX labelOffsetY boxStrokeColor boxFillColor textColor lineWidth fontName fontSize align vAlign shrinkToFit label borderSpec'''.split(),False),
            letterBoxes=IntentSpec('letterBoxes','content','''style x y count label labelFontName labelFontSize labelTextColor labelOffsetX labelOffsetY boxWidth boxGap boxHeight combHeight boxStrokeColor boxFillColor textColor lineWidth fontName fontSize'''.split(),False),
            amount=IntentSpec('letterBoxes','content','''style x y count label labelFontName labelFontSize labelTextColor labelOffsetX labelOffsetY boxWidth boxGap boxHeight combHeight boxStrokeColor boxFillColor textColor lineWidth fontName fontSize'''.split(),False),
            )
    intentSpecs['letterBox'] = intentSpecs['letterBoxes']

def _test():
    import doctest
    return doctest.testmod()

if __name__ == "__main__":
    _test()
